# Tcl package index file, version 1.1
# This file is generated by the "pkg_mkIndex" command
# and sourced either when an application starts up or
# by a "package unknown" script.  It invokes the
# "package ifneeded" command to set up package-related
# information so that packages will be loaded automatically
# in response to "package require" commands.  When this
# script is sourced, the variable $dir must contain the
# full path name of this file's directory.

package ifneeded notebook 1.0 [list source [file join $dir notebook.tcl]]
package ifneeded mactabnotebook 2.0 [list source [file join $dir mactabnotebook.tcl]]
package ifneeded moviecontroller 1.0 [list source [file join $dir moviecontroller.tcl]]
package ifneeded tree 1.0 [list source [file join $dir tree.tcl]]
package ifneeded progressbar 1.3 [list source [file join $dir progressbar.tcl]]
package ifneeded ProgressWindow 1.0 [list source [file join $dir ProgressWindow.tcl]]
package ifneeded TkInteractor 1.0 [list source [file join $dir TkInteractorPackage.tcl]]
package ifneeded sha1pure 1.0 [list source [file join $dir sha1pure.tcl]]
package ifneeded chasearrows 1.0 [list source [file join $dir chasearrows.tcl]]
package ifneeded wizard 1.0 [list source [file join $dir wizard.tcl]]
package ifneeded entrypopup 1.0 [list source [file join $dir entrypopup.tcl]]
package ifneeded combobox 2.2 [list source [file join $dir combobox.tcl]]
package ifneeded fontselection 1.0 [list source [file join $dir fontselection.tcl]]
package ifneeded Pane 1.0 [list source [file join $dir pane.tcl]]
package ifneeded balloonhelp 1.0 [list source [file join $dir balloonhelp.tcl]]
package ifneeded can2svg 1.0 [list source [file join $dir can2svg.tcl]]
package ifneeded svg2can 1.0 [list source [file join $dir svg2can.tcl]]
package ifneeded svgwb2can 1.0 [list source [file join $dir svgwb2can.tcl]]
package ifneeded can2svgwb 1.0 [list source [file join $dir can2svgwb.tcl]]
package ifneeded MSSpeech 1.0 [list source [file join $dir MSSpeech.tcl]]
package ifneeded uriencode 1.0 [list source [file join $dir uriencode.tcl]]
package ifneeded putfile 1.0 [list source [file join $dir putfile.tcl]]
package ifneeded getfile 1.0 [list source [file join $dir getfile.tcl]]
package ifneeded httpex 1.0 [list source [file join $dir httpex.tcl]]
package ifneeded tinyfileutils 1.0 [list source [file join $dir tinyfileutils.tcl]]
package ifneeded undo 0.1 [list source [file join $dir undo.tcl]]
package ifneeded entrycomp 0.2 [list source [file join $dir entrycomp.tcl]]
package ifneeded tinyhttpd 1.0 [list source [file join $dir tinyhttpd.tcl]]
package ifneeded buttontray 1.0 [list source [file join $dir buttontray.tcl]]
package ifneeded headlabel 0.1 [list source [file join $dir headlabel.tcl]]
package ifneeded hooks 1.0 [list source [file join $dir hooks.tcl]]
package ifneeded pipes 1.0 [list source [file join $dir pipes.tcl]]
package ifneeded component 1.0 [list source [file join $dir component.tcl]]
package ifneeded alertbox 1.0 [list source [file join $dir alertbox.tcl]]
package ifneeded colorutils 1.0 [list source [file join $dir colorutils.tcl]]
package ifneeded base64 1.0 [list source [file join $dir base64.tcl]]
package ifneeded md5 2.0.0 [list source [file join $dir md5x.tcl]]
package ifneeded wavelabel 1.0 [list source [file join $dir wavelabel.tcl]]
