#  WBIcons.tcl ---
#  
#      This file is part of The Coccinella application. It contains
#      Mime (base 64) coded icons for the tool buttons and other icons.
#      
#  Copyright (c) 2002-2003  Mats Bengtsson
#  
# $Id: WBIcons.tcl,v 1.1 2004/06/06 06:41:31 matben Exp $

package provide WBIcons 1.0

 
set ::WB::iconsPreloaded(off20) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEcTDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui8YkTNczB+x8z3+2DoBALBpPBE9QlyQNlbnN09lcRQ/TUXazlFZBW9M1HPqKW0Jz
GYpmlsJdrDrN/vjuO3PcdOzrrzdWbYEde4SGgYg3ijQwfo+QJ3iTlD0RADs=
}]
 
set ::WB::iconsPreloaded(off01) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEdjDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui5aA3JJAnsJ4rpM8GmEInOEIOyOntkRumJ7gEnoAOKvUjRTr63qzW+xwaB3niCvl
82ptXrVqbbPzht+OyfvHWwqHzn1xIXUfflE2MXmJgXqLdIKOMGOTlJWWX5iZmREAOw==
}]
 
set ::WB::iconsPreloaded(off40) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEcTDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui8Yk/AF4rgOgHQKEEy44/Pg6QMIhp1w2O0fohrkyznrFUtQjfI620GztehN/yT/v
uYUFythWdwo8Na9Hu3ctGJyjaWFwgFx/gyaFg3SJInyNjo+QO5KTlDkRADs=
}]
 
set ::WB::iconsPreloaded(off41) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEfjDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui8YkTNfz/dkl4P8+D2+UqfgIneGnSAEgAMhVLoRBPKMbJYc5cV6xpumGK/FCweFW
h5wwE9DpEPBnPYOUXmv9O8Lr9UdwUmpjc4ElWm+KiilaOllij3GSQpGPjpcii5ucnYqGoKGh
EQA7
}]
 
set ::WB::iconsPreloaded(off30)  [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEfTDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui5ZADZBwCiD8DeakHQ8BIPxmo1qPYPwAQZmErelEfjJLqucZkixx1g6WyAS3rl5y
icsZF7XVc0hYTLE3trc9jK+V93JiejKBbX+EODEmfIorjI2LhZAbd5MwTJiZmpuaeZ6foKAR
ADs=
}]
 
set ::WB::iconsPreloaded(off31) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEfzDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui7oAggCEB8f1nXc7VQ/3W81Sw6KxhbQRP8FSEhQdJZVLoROblQK+BO6mGnLiSGRO
JnE7j9KbDOLkph47cnoYfVfPwWJdIQl6gYJqX0SGhythi0B9MTqRkoyVkBqOmpucnYmfoKGi
nxEAOw==
}]
 
set ::WB::iconsPreloaded(off10) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEUzDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui8YkTNfz/dn6nverH9AkBPIOgKRyyQR4jiaCdEqtdqBDrLHY03a5Om8YfBOXyTRz
Gh2DVd9weHNOp0cAADs=
}]
 
set ::WB::iconsPreloaded(off11) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEZDDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui8YkTNfz/dkugLc9H4inAhiHuZTx6CEun9BnM0moWq/PaoeoNGq3yZKXsAuTAN8p
0EVGrnXgN3zDnZvM8Ppcn8frYFeBgoNWUYaHhxEAOw==
}]
 
set ::WB::iconsPreloaded(on20) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEc3DISau9A+jNu+9JZoxkaZ5l
CKBsa6hunIqyDNfxjbf6+HmongFAKBqPBOJJqGw1Z6vSEzUdMQmu6otG0kqxpmsWDDV5u2SS
2Jm2cn1tc5z548zfI6S+uMTvwn5/aoGCW1GFgIeIg4qLho4zCZKTlJWWlhEAOw==
}]
 
set ::WB::iconsPreloaded(on01) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEfnDISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1HN+nqKe3XKsT9OkISJvQBSDwloadqamS4oxTKzXrEpI8raUGiWySx84X
dpqOts3htendNtRVUNZWKa9yikBMY4BWQ3ddfSuHiIE8NImOao2RcYWUV5aXeAMJnZ6foKGh
EQA7
}]
 
set ::WB::iconsPreloaded(on40) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEeXDISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1recqmbo8k4ZA2BCHQBELQDQYhYTkbxV1Ik1VnNJVZO62rW6WBmaJv9Pl
1RZcrdll1RkdC7/J6d5d1rZudCdxWERNN32AWnmIiXWLUo2OiZF8GZNJCZiZmpucnBEAOw==
}]
 
set ::WB::iconsPreloaded(on41) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEgnDISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1ree6zfeyX6uEE80AiaRSQyimXEhlEoAANF/GoXSKqF5VQkNUSvV+TcLx
sms9o7MmNdfs4o3YBDcYXu6y6UF8fn5MelhPYh95Nz95jo47cEB1kpNOMZaUiJmHnJqeLSFb
o6SlCREAOw==
}]
 
set ::WB::iconsPreloaded(on30) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEf3DISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1zNm5DCDAmKzX+gERAMJQpDMmlS7iCpBIbKBRposaJGBbUhM10eVpWVyk
15xCV8ts4ZSs/mbbU6f9Lldx1jdhV3t8RU83KoKAiCdnK4Q0joyFk5F4lWAZmJabS1WfoKGi
oxEAOw==
}]
 
set ::WB::iconsPreloaded(on11) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEaHDISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1ree6zfey3yxGE/VSQaNuk0TeOK4c6dPiEa7YLOeKUz413K7TBiZEveRw
FW0zN4nAd7w4nr/YdpMwf8LnUXxrcIF6AwmHiImKi4sRADs=
}]
 
set ::WB::iconsPreloaded(on31) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEh3DISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1bQAIAhCt3I3nA+JEOuLP+EIOez+XUKYkMJspW1U2bVWv2Nhsy3W6OFaa
maWMqrPsRMLnLsNXgDlBAz7eTXkIe0s2XTuCfH1+Yip5g2k3U2iQkWsGVpQ6hjpBlpydjJ92
ooUDcqeoqaqrEQA7
}]
 
set ::WB::iconsPreloaded(on10) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEWnDISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1ree6zfeyH7AlHL5ExiAySVySPs2UiUCtWq8E3JJ5lHJVRWY4OTaWh2dg
urfebb+nDLw7Bw8S+Lx+z+dHAAA7
}]
 
set ::WB::iconsPreloaded(off00) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEZjDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui8YkPAK0DQK4q3+8XuoXCspau6BwROwoeaUmJ0ioEmozYBVwjWZvXSyyxPWOSWVx
LA2S7sIfd1tNo9fb3/sqrzfx9TBWgoOEhU+HiIlBEQA7
}]
 
set ::WB::iconsPreloaded(on21) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEfnDISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1redtd/MqDWE4TMmABiGBZWwhAUNXzAWELmVTnGgFtXW1zeQVO1ZVyz30
aRtUc9XPmdXJDma5X/DdFE7O6X12G0RHdS1EUTSGOoWBjHqPNkiRLxmUinuXZgMJnZ6foKGh
EQA7
}]
 
set ::WB::iconsPreloaded(off21) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEejDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui8YkTAI4XtofTvw/HYjX8RFCgOGs+AMlP8QNoDl6dqKHaUl7XUqPN/DKmxU7zSYy
t4rGZsNtctZaXI9bPfqc2sU7c1N8fSlAgh5uNHeJNXKLaX6OipGHIoWWl5hAgJucnZ0RADs=
}]
 
set ::WB::iconsPreloaded(on00) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEa3DISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1redsrKu8Vep3Erk2xKCK81OaOjtjy3NzGjaELMFmBWS90aEMXJXOyDTz
eMtVr9viM1tmbc1ddaIpr+cn3XovGYFpPoQtIQmKi4yNjo4RADs=
}]
 
set ::WB::iconsPreloaded(off50) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Ojo6Pj4+P//////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEnFDJSau1QJ3Nu/9eBo7kQYhl
yp2aqrIpQbzoKMtx/ckIgpMwGwEB+Nl0HlnGCAqCABmFEnBsPSW+IZSaRHKiCEWxt+1aO2Ax
sXeCdpxfrILMbhdXyHRvP4XONnAHGWF7dUVbfyZeg4VKfYmBG4xZRTOIeGdohj9+kF5fjomV
b59fo29mMYlApS6kma47rbGKGje3uLm6W7y9vr+8EQA7
}]
 
set ::WB::iconsPreloaded(on50) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEn3DISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wTBDzK8r3nZu11g2B8OWCrCHASNsJCYAEU4ZcaRJSaIzqZEUTRQJi46qavojE
kkhmmdHqMVG7VVXh7DkdZ+/iiXocfEB+YGCAYUsbgyddBlGIiTeLOilWh5I4lH2WVoFGGlNm
Z5NMS4yNnV6nSS2jKlNNqj+ydbSyt7QouT8hWL/AwcLDEQA7
}]
 
set ::WB::iconsPreloaded(off51) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEfDDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui5YAkMI0cdojPgK6zQ7k+wAJnmGn2Dn2eMJZM0gCrqQbZ+7a4mi30e7hCzZJrS60
WTyOqZmnVI26/sRBc+QSGw2RuU9Zf3tse4NJfEt6N4kxhI4lTJCPOpWWl5iXNZucnZ6fNREA
Ow==
}]
 
set ::WB::iconsPreloaded(on51) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEgnDISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1nN65CxCEnazH+hFMwZiLqDKyki2mwekK4kRFoO33wja126OqRw13d2Xz
2Mu9mdomsvstjnsN1m2amG9pwGdKKn1vaYEteX91S3crSYA0jY6LNlJzV0KXjIKam52RCaGi
o6SlpREAOw==
}]
 
set ::WB::iconsPreloaded(off60) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEdzDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui8YkHAI47tocQPxAXYnX+5wAtVmPADqNiEWQr6lcjpDGavQz9UCt0mwrLGV2vj1s
yLzSro5rr/sM33Tb4yH8vkE/j2x9czU/cnk0YohJh4p4jVQaQJKTlJU5l5iZmpcRADs=
}]
 
set ::WB::iconsPreloaded(on60) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEfnDISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1reemp5s8AGFITN14hhiLYJQhhS4mTcRStqAuZJIgs+Ko1S63pd12s+Cw
C/tKr5rvcRuuGtLZqnKdacUD3VF8Zl90MnZ+J4A2RGSKP4Rej42FknkZlVORmJYJnZ6foKGh
EQA7
}]
 
set ::WB::iconsPreloaded(off61) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEeTDJSau1IJ3Nu/9eBo7kQYhl
yp2aqrLui8YkHHJAnpd2t+vA2gwHJBhPgFGPmDP6CKDlBpj0QD/SA7WaypqOzu6Qhh2PtuWW
inpdmaefovWt5dZ17Y13ukXmTXR8eFAAf3srYCtzamRpjUqBZIeSImCWl5hHW5ucnBEAOw==
}]
 
set ::WB::iconsPreloaded(on61) [image create photo -data {
R0lGODlhHAAcAOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+P//////////////
/////////yH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAHAAcAAAEe3DISau9A+jNu+9J9o2jYYSa
qa5su6KAK8/wbLP1redm3HM3niH1sQk7hCQhJTtukisAoSmKAllTl/O6q66USt3JK6aRZR7t
Gc3JvtbDVrsl/FmfdPjQs5TmmXJtU344eiZgbm4qdWWFgI1/PpB/k1SSlW8JmpucnZ6eEQA7
}]

# Other...

set ::WB::iconsPreloaded(barvert) [image create photo -data {
    R0lGODlhCwAyAPcAAP//////////////////////////////////////////////////////
    //////////////////////////////////////e1vfe1vfe1vfe1vfe1vfe1vfe1vfe1vfe1
    vfe1vfe1vfe1vfe1vfe1vfe1vfe1vfe1vfe1vfe1vfe1vfe1ve+cre+cre+cre+cre+cre+c
    re+cre+cre+cre+cre+cre+cre+cre+cre+cre+cre+cre+cre+cre+cre+cre9zhO9zhO9z
    hO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9zhO9z
    hOdae+dae+dae+dae+dae+dae+dae+dae+dae+dae+dae+dae+dae+dae+dae+dae+dae+da
    e+dae+dae+dae+dKY+dKY+dKY+dKY+dKY+dKY+dKY+dKY+dKY+dKY+dKY+dKY+dKY+dKY+dK
    Y+dKY+dKY+dKY+dKY+dKY+dKY94xUt4xUt4xUt4xUt4xUt4xUt4xUt4xUt4xUt4xUt4xUt4x
    Ut4xUt4xUt4xUt4xUt4xUt4xUt4xUt4xUt4xUtYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYA
    IdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIc7Ozs7Ozs7Ozs7Ozs7Ozs7O
    zs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7OzpycnJycnJyc
    nJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJyc
    nISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISE
    hISEhISEhISEhFJSUlJSUlJSUlJSUlJSUlJSUlJSUlJSUlJSUlJSUlJSUlJSUlJSUlJSUlJS
    UlJSUlJSUlJSUlJSUlJSUlJSUlJSUgAAACwAAAAACwAyAAAIfABVCRxIsKBBgRdYBLHC0Moa
    QNQEVppIsRK1XwMrWsRIkOLFgxsPilQFgGDJgukIphxZ8KRAlwNXCpTJ8qVJgzRp1oQJc6bK
    mjcH9lSVE6jQoD9jGrV5FGVSozxxPgUa1anSpVWn6mSZ9apPrEi9El1KMuzXs1TNjhU7MiAA
    Ow==
}]

set ::WB::iconsPreloaded(barvert) [image create photo -data {
    R0lGODdhCwA2ALMAAP////e1ve+cre9zhOdae+dKY94xUtYAIdDQ0M7Ozpyc
    nISEhFJSUgAAAAAAAAAAACwAAAAACwA2AAAEYjDJSauVQQzCSTGLdIzksShT
    aaIUeV7rJScAVVcMlc/VLfmTnUTI+9ksRGIRCBzqisdJM5GESqPPoNV4xWWt
    TOQXGvZqt+Wxkpc+O9FYN3VLi7/vZPtcvtRXEwiBgoOEhQgRADs=
}]

set ::WB::iconsPreloaded(barhoriz) [image create photo -data {
    R0lGODdhQQALAOcAAP//////////////////////////////////////////////////////
    ////////////////////////////////////////////////////////////////////////
    /////++Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+U
    pe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe+Upe97lO97lO97lO97
    lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97
    lO97lO97lO97lO97lO97lO97lO97lO97lO97lO97lNYAIdYAIdYAIdYAIdYAIdYAIdYAIdYA
    IdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYAIdYA
    IdYAIdYAIdYAIdYAIdYAIdYAIc7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7O
    zs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7Ozs7O
    zs7Ozs7OzpycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJyc
    nJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnJycnISEhISEhISE
    hISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISE
    hISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhISEhFJSUv//////////////////////
    ////////////////////////////////////////////////////////////////////////
    /////////////////////////////////ywAAAAAQQALAAAIdwABCRxIsKDBgwgTKlzIsKHD
    hw9DIAQgkCIgixgrary4MSPHjx4tHgwBBgzCbwJRAlLJMqXLlS9bwpwpU2VBkiVNQtzJc2TO
    nD2D8gTy86fQow5/KV26dGDIjlBBRn0q9SCoq1ivEqwZsytNr1y/Ih1LtqxZhAEBADs=
}]

set ::WB::iconsPreloaded(imcolor) [image create photo -data {
    R0lGODlhOAA0AOMAAAAAAHBwcHh4eIiIiJCQkKCgoMDAwNDQ0NjY2Pj4+AAAAAAAAAAAAAAA
    AAAAAAAAACH+Dk1hZGUgd2l0aCBHSU1QACwAAAAAOAA0AAAE/jDJSau9OGuQjv9gKI5kWXJm
    qq4jgbJw3L5yHbvdR+x87//AnYigwoFcgKRyyWw2czoA0WSMaq6VA/SA5JWqHhx2rI0yp0Ma
    lzPGls086Wwrbmve0akLbT2y7Rl4On58a3R/gBeCM4VghokYiyE7UnyOdZBZW5MAnZ1ol4iZ
    E5J+n36HoxalhJypqhSsR2mvsBKyVGqYsLhfuqKqWnYAfVa2t5smxGG/x8htLsXMzh120dOo
    zsJj14bZx9tY3aHU4VfjzdrWy97G4OvSj+/Q7KE2JPDYOtSJ6Mkq1fj5qxGQ2kAZBZ0dDANE
    RMJjC9d4OhXioa2ISCYWKsMvAcaJXRRBWIT1EaQPDyNVldQ4YUrKUSs9EWiJ0tzFerqA0OQo
    EOc/LhZc2iTpk8TMCkI7RjwStGaQp1ApxTvS8ADIq1izXp16r+scr2CNqglL1kjUs2jTxtHK
    tq3btwAiAAA7
}]

#-------------------------------------------------------------------------------
