# Tcl package index file, version 1.1
# This file is generated by the "pkg_mkIndex" command
# and sourced either when an application starts up or
# by a "package unknown" script.  It invokes the
# "package ifneeded" command to set up package-related
# information so that packages will be loaded automatically
# in response to "package require" commands.  When this
# script is sourced, the variable $dir must contain the
# full path name of this file's directory.

package ifneeded CanvasCmd 1.0 [list source [file join $dir CanvasCmd.tcl]]
package ifneeded CanvasCutCopyPaste 1.0 [list source [file join $dir CanvasCutCopyPaste.tcl]]
package ifneeded CanvasDraw 1.0 [list source [file join $dir CanvasDraw.tcl]]
package ifneeded CanvasFile 1.0 [list source [file join $dir CanvasFile.tcl]]
package ifneeded CanvasText 1.0 [list source [file join $dir CanvasText.tcl]]
package ifneeded CanvasUtils 1.0 [list source [file join $dir CanvasUtils.tcl]]
package ifneeded FilePrefs 1.0 [list source [file join $dir FilePrefs.tcl]]
package ifneeded GetFileIface 1.0 [list source [file join $dir GetFileIface.tcl]]
package ifneeded Import 1.0 [list source [file join $dir Import.tcl]]
package ifneeded ItemInspector 1.0 [list source [file join $dir ItemInspector.tcl]]
package ifneeded Multicast 1.0 [list source [file join $dir Multicast.tcl]]
package ifneeded Plugins 1.0 [list source [file join $dir Plugins.tcl]]
package ifneeded PutFileIface 1.0 [list source [file join $dir PutFileIface.tcl]]
package ifneeded WBIcons 1.0 [list source [file join $dir WBIcons.tcl]]
package ifneeded WBPrefs 1.0 [list source [file join $dir WBPrefs.tcl]]
package ifneeded Whiteboard 1.0 [list source [file join $dir Whiteboard.tcl]]

