
#  Base64Icons.tcl ---
#  
#      This file is part of The Coccinella application. It contains
#      Mime (base 64) coded icons for the tool buttons and other icons.
#      
#  Copyright (c) 2002-2003  Mats Bengtsson
#  
#  See the README file for license, bugs etc.
#  
# $Id: Base64Icons.tcl,v 1.10 2004/09/30 12:43:07 matben Exp $
 
### First all the tool buttons, both on and off states. ########################

# Make sure we have the namespace.
namespace eval ::UI:: {}


### Mac Head icons. ############################################################

set ::UI::icons(machead) [image create photo -data {
    R0lGODlhEAAMAPMAAP//////zv/OnN7e3s6cYwAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQAAwAAAQxcJRJ57hYhs2D
    KNlVeEJZEmA4mp+AhhpbuGm2sjR8s69KmhTdb/aBxYAWIa6nqyRhEQA7
}]
set ::UI::icons(macheadaway) [image create photo -data {
    R0lGODlhEAAOAPMAAP//////zv/OnN7e3s7O/86cY2NjzjExYwAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQAA4AAARHcMgpkb2IUhS6
    F5k2cEJZFqFGloiATtgKumkdmGcWYyYvECYgqAUauQyugqGAGuqOTCWT+GpJ
    mUvUpQLFTlWuQ1T8Uu0soggAOw==
}]
set ::UI::icons(macheadgray) [image create photo -data {
    R0lGODlhEAAMAPMAAP///97e3sbGxoSEhAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAEALAAAAAAQAAwAAAQxMJBJZ7hYgs2B
    INlFeEJZDmA4miYaauYkuCrJ0tnKzmlul5SXjtXzxYLCH+8lqliYEQA7
}]
set ::UI::icons(macheadinv) [image create photo -data {
    R0lGODlhEAAMAPEAAP///97e3gAAAAAAACH5BAEAAAEALAAAAAAQAAwAAAIl
    jI4XywYPn2hLxDipxZI6DmTNdoUeeYkMKikaqK6p+5bxmtBUAQA7
}]
set ::UI::icons(macheadsleep) [image create photo -data {
    R0lGODlhEAAPAPMAAP//////zv/OnN7e3s6cYwAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQAA8AAAQ+cEhZapk4U03r
    vNzWeZoVnoOlllbgBoIAUkFRxzgxpzcuCzqMreD7BT+wYuxIUy53NicQGiBY
    r9ddSsXFRAAAOw==
}]
set ::UI::icons(macheadtalk) [image create photo -data {
    R0lGODlhEAAOAPMAAP//////zv/OnN7e3s6cYwAAYwAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQAA4AAAQ/cJhJ57hYhs2D
    MNlleEJZEmA4mp+AhhpruGm2svRQ2CT76pibzLDbiXqVS/FoqimBsaYFaBQ2
    qbbK9AnrGi8RADs=
}]
set ::UI::icons(macheadunav) [image create photo -data {
    R0lGODlhEAANAPMAAP//////zv/Ozv/OnP9jY/8xAN7e3t4AAM6cY70AAK0A
    AIwAAHMAAFJSUlIAAAAAACH5BAEAAAYALAAAAAAQAA0AAARV0Jgjq6R2msc7
    Z9YhPMQSnImgPGC1LooyHIXSPNkGK/Sz4pnHLsZCAC2PwCGmYAyMuUeixlRA
    g4QYg8jJHITNgW+lCBGEzoGY07QwHg61/JN7e+iWCAA7
}]
set ::UI::icons(questmark) [image create photo -data {
    R0lGODlhEAANAPMAAP////9jY97e3q0AAFIAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAIALAAAAAAQAA0AAAQmUIgQxpA4y2qt
    zpzUfeKFeeRAZCoptBvsxutM22eNv/ruxz+ZKwIAOw==
}]
set ::UI::icons(dndim) [image create photo -data {
    R0lGODdhEAALALMAAP//////zv/Ozv/OnM6cnM6cY85jY5xjY5xjMZwxMQAAAAAAAAAAAAAA
    AAAAAAAAACwAAAAAEAALAAAERVDJSasMQoxBevmgJEhEoRiIYipIqoyDQkimGbvxPKv8LcUm
    HUnGGh6IO6FhmJoVUjmelJAqHGOKwaqQ6HZbLcNnZylLIgA7
}]
set ::UI::icons(naim) [image create photo -data {
    R0lGODdhEAALALMAAP///87//5z//5zO/2PO/2Oc/2OczjGczjFjzjFjnABjnAAAAAAAAAAA
    AAAAAAAAACwAAAAAEAALAAAESHDJSasEIAQxOiFFSFzBEozDgUhJK22SsAzJPMy1ec/4vNIx
    lAQ4IQ4KvF5qIVjNDMofVACdHZTXo6lKUHgVhFYLMTBULWhJBAA7
}]

# With whiteboard icon.
set ::UI::icons(macheadwb) [image create photo -data {
    R0lGODlhEAAOAPMAAP//////zv/OnN7e3s6cY29vbwAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQAA4AAARCcJhJ57hYhs2D
    MNlleEJZEmA4mp+AhhpruEZh3ytLAzxf5CxUzwcsUYaAH6n1QSaLR6RSJ5RK
    KhNnMlTQ+ri38C0CADs=
}]
set ::UI::icons(macheadtalkwb) [image create photo -data {
    R0lGODlhEAAOAPMAAP//////zv/OnN7e3s6cY29vbwAAYwAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQAA4AAARKcJxJ57hYhs2D
    ONl1eEJZEmA4mp+Ahhp7uEdh3ytLAwbgFzkWytcDAEmyQ694NFWIPGPw4Ksu
    pcjPBNpr6qhXo6Sytf5Ct5shHQEAOw==
}]
set ::UI::icons(macheadawaywb) [image create photo -data {
    R0lGODlhEAAOAPMAAP//////zv/OnN7e3s7O/86cY29vb2NjzjExYwAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQAA4AAARNcMgpk72J0hS6
    F5k2cEJZFqFGlomATtgKumkdmGcWYyYvECYgqAUytAoHF7KAGiaMrgNz2aRB
    qVQM4CpldhOA8PaImJZR4u1oZzG43xEAOw==
}]
set ::UI::icons(macheadinvwb) [image create photo -data {
    R0lGODlhEAAOAPMAAP////7+/vr6+vb29t7e3nV1dW9vbwAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAQALAAAAAAQAA4AAAQ7kJxJJ7lYgs33
    ydfRdR8okh6oocBhvPA5tkNnyOMR2Lg3jQXcSTe65X7FnmdmkFQmAibIMOM0
    M7BsNgIAOw==
}]
set ::UI::icons(macheadunavwb) [image create photo -data {
    R0lGODlhEAAOAPcAAP//////zv/Ozv/OnP9jY/8xAN7e3t4AAM6cY70AAK0A
    AIwAAHMAAG9vb1JSUlIAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAYALAAAAAAQAA4AAAiJAA0YOCCwoECC
    BgcagMCwIUMGBg8IgEBgQYCLCQQogACx4MYFChQMOFBAgQMICReCVEASwkYI
    DWI2EAhhZUiOCCAA2DlzYYADIRUwGJBzJ4CeEBKUDKqgKE+aBEIyuMnQ6MwD
    NYUOcLlRgdUDBGoOHbCVodCnDCA8IMv2oVWBaR26fZtSpt2YAQEAOw==
}]
set ::UI::icons(macheadsleepwb) [image create photo -data {
    R0lGODlhEAAPAPMAAP//////zv/OnN7e3s6cY29vbwAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQAA8AAAREcEhpqpk4U03r
    vNzWeZoVnoOlFmwxGkEcCIIBAC6s0zxh46mAgUcb+m45IbEmOAJhy94vGaU5
    qdVrkMDtdqcvVaVViAAAOw==
}]
set ::UI::icons(questmarkwb) [image create photo -data {
    R0lGODlhEAAOAPMAAP////9jY97e3q0AAG9vb1IAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAIALAAAAAAQAA4AAAQ9UIgQxpA4y2qt
    zpzUfeKFeeRQYIRKvPCrYlUB3DhAgGqe75ueDwcUuIbEU4GAvBUlzGaxEkUW
    j1JNbPuKAAA7
}]
set ::UI::icons(macheadgraywb) [image create photo -data {
    R0lGODlhEAAOAPMAAP///97e3sbGxoSEhG9vbwAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAEALAAAAAAQAA4AAAQ/MJRJZ7hYgs2B
    KNlVeEJZDmA4miYaauYkoERtr+w8dgTOoh1AjxSbBIe5T1DoqyyRvx1PUjE+
    Q4Qlh4C1eW0RADs=
}]

### Foreign IM systems icons. ##################################################

set ::UI::icons(aim_away) [image create photo -data {
R0lGODlhEAAQAPUAAP//AP+4SP7LZv31tvvUt/Pz8+vp5eThzODg4N/fg9ra
2tbNu87n/8ujR7y8AKunn6bU9qainJ6enp2ckJmDZpjL/pVtMo2Zo4egsYLD
+YCDTH2fzX2BAXFxcG+QkW1gLGZmZmBLAlAsAEay6zqn8zMzMDIuBjFfKB+W
8B87fRsnGg+Q9g8PDQ0rQAVkmAFnywAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAsALAAAAAAQABAAAAan
wIVwoTBIWEiQQTEUFhkbmKVhYWGWwyKElCoJBIlPC4JdGCCoVfcrEKPIRAZp
tXrBPg0N67UiMZ4VgRcRDzAwLS0uKxkGIY6Pjg8sJiUsKyONAJqbIQNfDZYo
mRykHAAhFBYfJqGZm5ohHTAsipejpacEVbWMIa+wBAG7fgq+paYhGg0BHyxw
vr8hFybLKlgfDgPa2hweFyUiLAhCEycfkMkHLCoFQ0EAOw==
}]
 
set ::UI::icons(aim_dnd) [image create photo -data {
R0lGODlhEAAQAPUAAP//AP+4SP7LZv31tvvUt/Pz8+vp5eThzODg4N/fg9ra
2tbNu87n/8ujR7y8AKunn6bU9qainJ6enp2ckJmDZpjL/pVtMo2Zo4egsYLD
+YCDTH2fzX2BAXFxcG+QkW1gLGZmZmBLAlAsAEay6zqn8zMzMDIuBjFfKB+W
8B87fRsnGg+Q9g8PDQ0rQAVkmAFnywAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAsALAAAAAAQABAAAAan
wIVwoTBIWEiQQTEUFhkbmKVhYWGWwyKElCoJBIlPC4JdGCCoVfcrEKPIRAZp
tXrBPg0N67UiMZ4VgRcRDzAwLS0uKxkGIY6Pjg8sJiUsKyONAJqbIQNfDZYo
mRykHAAhFBYfJqGZm5ohHTAsipejpacEVbWMIa+wBAG7fgq+paYhGg0BHyxw
vr8hFybLKlgfDgPa2hweFyUiLAhCEycfkMkHLCoFQ0EAOw==
}]
 
set ::UI::icons(aim_offline) [image create photo -data {
R0lGODlhEAAQAPUAAP6pP/vNY/i3UNbNu9aoTNScMr+9tr+bXbi1r7Opo7Cw
r6yrq6h9QaSinaOalJycnJmXkpKSkpJhJZGIgoqQjImJiYhycIWFhICAgHJv
dGxaA2hlZGVZVF1cXFBNS0BAQDszBDEyMiEhIQwPDgsKCAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQABAAAAaY
wIFQaEhEKhXHwjBsDhAY0OEgwjCdwkSIEAgcQgmscNHhekMI8cAQJRBAnSu2
YfGMNg/1oNE5BBgiIh4kF2FNUV2JASB4ThscXQySH4SGWR4MAgISJBsOWAgV
HwSaAm8bCk0GDRmZBAykp6lCDiAMAAIaFyKZBSMVWRuxHsATJJnEQqISEiIT
QxUkg89CEBUUE9RC2RPAA0EAOw==
}]
 
set ::UI::icons(aim_online) [image create photo -data {
R0lGODlhEAAQAPUAAPvPaPqoVfLz8uvp5eDg4Nra2tbNu82gUsrl9p7R9J6e
npm2xJOCOI6fqobG+HtRCm6n3G5ubmdnZ1VVVUur70VLA0MrD0BAQDCC0y5H
Yy2Y6SsYAicqKRscGhgpGBdLjRMpQxEQEA+P8gxaqgkKDwKLuAIaKAFoygFY
egBCUwAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAYALAAAAAAQABAAAAa0
QIPQUBgoOiSSZFAYCosIiIpxYJAazGExoUlZAOAHKJE1DLiizxc8AYnIRIRG
JDqpHlXSSaRBFB10IiYWFyoqJiAoIg4DAxR0KBxgFSEWFiYiFI1zggxgnwck
fJt0JAcABw8MDxwho44iJSEBqIUkipmNgCAXAbQBViN0jAULIBsHwAy+wX0F
BSoVyQcb0b4PJnB3AAHUKRgmHr4dWYarIRmPGoMVKk1EhioRBQQE9CQdAkNB
ADs=
}]
 
set ::UI::icons(aim_xa) [image create photo -data {
R0lGODlhEAAQAPUAAP//AP+4SP7LZv31tvvUt/Pz8+vp5eThzODg4N/fg9ra
2tbNu87n/8ujR7y8AKunn6bU9qainJ6enp2ckJmDZpjL/pVtMo2Zo4egsYLD
+YCDTH2fzX2BAXFxcG+QkW1gLGZmZmBLAlAsAEay6zqn8zMzMDIuBjFfKB+W
8B87fRsnGg+Q9g8PDQ0rQAVkmAFnywAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAsALAAAAAAQABAAAAan
wIVwoTBIWEiQQTEUFhkbmKVhYWGWwyKElCoJBIlPC4JdGCCoVfcrEKPIRAZp
tXrBPg0N67UiMZ4VgRcRDzAwLS0uKxkGIY6Pjg8sJiUsKyONAJqbIQNfDZYo
mRykHAAhFBYfJqGZm5ohHTAsipejpacEVbWMIa+wBAG7fgq+paYhGg0BHyxw
vr8hFybLKlgfDgPa2hweFyUiLAhCEycfkMkHLCoFQ0EAOw==
}]

set ::UI::icons(icq_online) [image create photo -data {
R0lGODlhDAAMAPMAMf//////APb7AAD7AAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAMAAwAAAQrEEhCAJXYjkF2
zR3nXRYlnuaYciEXCBQZEwK8sfechzz7xR5PRoMalo6SCAA7
}]
 
set ::UI::icons(icq_away) [image create photo -data {
R0lGODlhDAAMAPMAMf////8A9vb79rS0tB0dHQD7AAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAEALAAAAAAMAAwAAAQvMEhjApXYlkJH
zcZmDAJxWZQhrB6Xkiv7xrTwqbXc1V7K0b0bRcbJWHoiI2opiQAAOw==
}]
 
set ::UI::icons(icq_chat) [image create photo -data {
R0lGODlhDAAMAPMAMf//////AAD7AAAAbQAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAMAAwAAAQuEEhCAJXYCkF2
zVQoTkRgnl1oDkPABuIpw9N6tiG3vmmnoxdLLJXRcDxFoVISAQA7
}]
 
set ::UI::icons(icq_dnd) [image create photo -data {
R0lGODlhDAAMAPMAMf////8A9vi0tPYAAAD7AAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAEALAAAAAAMAAwAAAQzMMhSApXYElJ2
zcUwdOJlDYAophyJrmK7iYJQfhRtD56bvjwKAQjDhYKljIbjUVoqpkAEADs=
}]
 
set ::UI::icons(icq_offline) [image create photo -data {
R0lGODlhDAAMAPMAMf//////APb7APYAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAMAAwAAAQrEEhCAJXYjkF2
zR3nXRYlnuaYciEXCBQZEwK8sfechzz7xR5PRoMalo6SCAA7
}]
 
set ::UI::icons(icq_xa) [image create photo -data {
R0lGODlhDAAMAPMAMf////8A9vb79vYAAAD7AAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAEALAAAAAAMAAwAAAQzMMhSApXYElJ2
zR3nXdZgniZXDIUgtK/ArlTcrhL73tSs268ZAYX6UELIjEbEUVoqpEAEADs=
}]
  
set ::UI::icons(msn_away) [image create photo -data {
R0lGODlhEAAQAPYAMf////cJ5NTQyNHvwc8vAMfun8TbkcDcybTOkbHxvafs
y6bAnqaxf6Taz6Olm52+p5e1pJa3rpW9vZCQkJAvAI+0soestIaifoWPg4Og
kYGoeYGioIGIhn+ltXqfpHedrHa8r3alvnaUcHGsvnGNXHCWrW66eWyXsmx/
aWuJg2dvc2am3maEe2SWZWC1lmCBkV2AkVyDpVyCnVu2jVt7g1mErlloS1dy
f1dwXlaAs1WU4FJyXlF5gE2MxEuM5kp/nEpmgUl1p0h4uEdbUUV/vUVnfkRv
lURel0NprkKOe0FjmEBwsEBohEBmo0BRVT9hgT9ahT9Sfj1ZUD1TdzxXfjtz
2DlpzjlXYzdo3jddsDFUhDBEfS9EcS5DUC1HjytmZytd1CtFXypNoCpLoChA
hCZIkiRUnyNOryJAgyI3TyE8Yx0/mh0xZxsvMxg5eBdDlBIoVwAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAEA
LAAAAAAQABAAAAetgAGCgwEaJDYOhIqCJgYMKG0qi4MzCQMQL1Bwk4IuCg0h
QlllnAFJICM9VVhnpSlfP0tWYG+lAU5qZGNrkwS+BAItIjgYiwQAAHEABAUI
F1IchMfIyssLGTRhQ4PTycgEDxswU0ZXgsdxcRPfER4yVEdEaQHdyMsVHzFK
XDpm5/bLKFgoUaPJliA+ogQ48IsABQkdTuRA4qXICjEBdjChwJHFDSBPtLhh
04UHmkAAOw==
}]
 
set ::UI::icons(msn_dnd) [image create photo -data {
R0lGODlhEAAQAPYAMf////cJ5NYYANTQyNHvwc8vAMfun8TbkcDcybTOkbHx
vafsy6bAnqaxf6Taz6Olm52+p5e1pJa3rpW9vZAvAI+0soestIaifoWPg4Og
kYGoeYGioIGIhn+ltXqfpHedrHa8r3alvnaUcHGsvnGNXHCWrW66eWyXsmx/
aWuJg2dvc2am3maEe2SWZWC1lmCBkV2AkVyDpVyCnVu2jVt7g1mErlloS1dy
f1dwXlaAs1WU4FJyXlF5gE2MxEuM5kp/nEpmgUl1p0h4uEdbUUV/vUVnfkRv
lURel0NprkKOe0FjmEBwsEBohEBmo0BRVT9hgT9ahT9Sfj1ZUD1TdzxXfjtz
2DlpzjlXYzdo3jddsDFUhDBEfS9EcS5DUC1HjytmZytd1CtFXypNoCpLoChA
hCZIkiRUnyNOryJAgyI3TyE8Yx0/mh0xZxsvMxg5eBdDlBIoVwAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAEA
LAAAAAAQABAAAAergAGCgwEaJDYPhIqCJgcNKG0qi4MzCgQRL1Bwk4IuCw4h
QlllnAFJICM9VVhnpSlfP0tWYG+lAU5qZGNrkwW+BQMtIjgYiwUCyAIFBgkX
UhyEx8nIBQwZNGFDgwUA3d4FEBswU0ZXgtze3QUSHjJUR0RpAdLTAhUfMUpc
Ombn0wUULJSo0WRLEB9RAiD4BXBChxM5kHgpskJMgB1MKGhkcQPIEy1u2HTh
gSYQADs=
}]
 
set ::UI::icons(msn_offline) [image create photo -data {
R0lGODlhEAAQAPUAMfcJ5OzS0urMzOjIyOW/v+G3t9+xsd2srNumptigoNWY
mNGPj8+Kis2Fhcy8tsuyqsuBgcl7e8Z0dMVwcMNtbcKhm8FnZ79jY75gYL5f
X71gYLtZWbpWVrdPT7ZMTLZLS7FISKtGRqlFRaVERKJCQpk/P5hcWpQ9PZFE
RJA7O406Oos5OYg4OIdIR4U3N4A1NXw6OnoyMnYxMXMvL24tLWgrK2QpKV0o
KF0mJlAhIUofHzcXFwAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAae
QIBwCNCMXiaiUmhRUEY72HLIIQwgoZVuKtwYEJMPycUFnCYUzmZzKqNqpFDn
EysDbrhY6mUXgjoqLX0MBwsXNIJTDhAHBYYkNjNLDwMLAgUMFyIvJTRKBAkN
AQYQGCAuJR45QwkJIwgCCREXHyUzEyusGSUGBAsSGR0kMSEUZBUSfwgHDRMY
HSEuKxEpACsrNDIkIyUpLC82ODgnMkEAOw==
}]
 
set ::UI::icons(msn_online) [image create photo -data {
R0lGODlhEAAQAPcAMf////cJ5Nvr3tXosdHvwcr0yMjlzcfun8Tft8TbkcP2
vMDcybzoybjZvbTOkbHxvbD0tbDcxKzPvKfsy6bosqbAnqaxf6TiqaTaz6Ol
m6HfuJ/S2J/GuZ7mr52+p5nisZe1pJbjpJa3rpW9vY+0sofTrIestIaifoWP
g4PPlYOgkYGoeYGioIGIhn+ltXu8vnqfpHedrHa8r3alvnaUcHGsvnGNXHCW
rW66eWyXsmx/aWuJg2dvc2am3maEe2SWZWGGcmC1lmCBkV6pfl2AkVyDpVyC
nVu2jVt7g1mErlloS1dyf1dwXlaAs1WU4FJyXlF5gE2MxEx80UuM5kp/nEpm
gUl1p0h4uEdbUUWKk0V/vUVnfkRvlURel0NprkKOe0FjmEBwsEBohEBmo0BR
VT+Eqj9hgT9ahT9Sfj5rTT1ZUD1TdzxXfjtz2DlpzjlXYzdo3jddsDFUhDBE
fS9EcS5DUC1HjytmZytd1CtbzStFXypNoCpLoChAhCZIkiRUnyNOryJAgyI3
TyE8YyA/fR0/mh0xZxsvMxg5eBdDlBIoVwAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAEALAAAAAAQABAAAAjAAAMIHBhghQ0l
GQgqFIgjgQUdh3gsHHjkAQEQQs4omigwyAQMM67E8cMxwBcZNaK0gQOo5I47
VMK4wZOoZAAyg/rwKWRT4A8aTFD0THHAwQk1LTgKSHFhQAUVSPRgWWhAQQkF
CDywILKGyxuFEDaUKNBABAwjbLpoETQwxIYyHQpIIBGjCBg6Tv4M/CAlCwUG
HEzcSDJmjpUpaAIseJFniIYII1zkaOLFzpYeewI8EUMoDRAfS6qYkYPIUB0o
gQICADs=
}]

set ::UI::icons(yahoo_away) [image create photo -data {
R0lGODlhEAAQAPMAAP//////AP8AAObe3L+/v39/AH8AAAAA/wAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQABAAAARDcMhJq71U6C3w
3B8HSuOgIaaWqULhFme1FUEdwKhM23Vx7YBa8LaiAQ4EwiHomyAQO17PQove
iNSbwVD1WJvel2ccAQA7
}]
 
set ::UI::icons(yahoo_dnd) [image create photo -data {
R0lGODlhEAAQAPMAAP//////AP8AAObe3L+/v39/AH8AAAAA/wAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQABAAAARDcMhJq71U6C3w
3B8HSuOgIaaWqULhFme1FUEdwKhM23Vx7YBa8LaiAQ4EwiHomyAQO17PQove
iNSbwVD1WJvel2ccAQA7
}]
 
set ::UI::icons(yahoo_offline) [image create photo -data {
R0lGODlhEAAQAPMAAP////8AAObe3L+/v39/f39/AAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAIALAAAAAAQABAAAAREUMhJq7006B3w
3B8HSqOgGaaWqULhFmelFgNQw6hM23xx7bUgwBcqDHeuAVFiMAB5N4tRGFxS
pgSCzVqZbj2SFxdMiQAAOw==
}]
 
set ::UI::icons(yahoo_online) [image create photo -data {
R0lGODlhEAAQAPMAAP//////AP8AAObe3L+/v39/AH8AAAAA/wAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQABAAAARDcMhJq71U6C3w
3B8HSuOgIaaWqULhFme1FUEdwKhM23Vx7YBa8LaiAQ4EwiHomyAQO17PQove
iNSbwVD1WJvel2ccAQA7
}]

set ::UI::icons(yahoo_xa) [image create photo -data {
R0lGODlhEAAQAPMAAP//////AP8AAObe3L+/v39/AH8AAAAA/wAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAACH5BAEAAAMALAAAAAAQABAAAARDcMhJq71U6C3w
3B8HSuOgIaaWqULhFme1FUEdwKhM23Vx7YBa8LaiAQ4EwiHomyAQO17PQove
iNSbwVD1WJvel2ccAQA7
}]

set ::UI::icons(gadugadu_away) [image create photo -data {
R0lGODlhEAAQAPUAMf//////AP/3AP/vAP/nAP/eAP/WAP8AAPfWAO/OAOfG
AN7GAM61AM6tAL2lALWcAKWMAJyEAJSEAISEhAD//wD3/wDr/wDb9wDO7wDO
5wDG3gC50gCtxgCnugCYsQCMoACElAB7jABtewBmZgBSYwAAmQAAjAAAhAAA
ewAAZgAAWgAAUgAASgAAQgAAMwAAKQAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAaX
QIBwCDgQj8jDBDk0NpdCJ9F4OAiqRejxEAgICgklkxuAEWAKhnjrHRgUDce6
WBWYCIhF4wGJVKUlJRQneg8wEicoKRNLghQUFicwhycWGBsqAI6QFpEnJxeX
HSsAJhUVFqEZqxkaGx4gIi2VGBkbHBwduR4fISIuQikbHbwgxrEiIyTAACkd
vS3RySMuLy9ELNFD1ddEQQA7
}]
 
set ::UI::icons(gadugadu_invisible) [image create photo -data {
R0lGODlhEAAQAPQAMf////8A9v8AAPf39+/v7+fn597e3tbW1s7OzsbGxrW1
ta2traWlpZycnJSUlIyMjISEhIQAAHt7ewAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAEALAAAAAAQABAAAAVgYCCOQUSe
qACho0BGUouakTAI9JoCwGAgKlaENyFMEoqIjiToEQyJBUMQE9UiA0LhEG08
IFem4bhoTCAqCaRaQiQm8LWSJZA61sEZWPDAlZYieX9VggGAhSwjMIksgCMh
ADs=
}]
 
set ::UI::icons(gadugadu_offline) [image create photo -data {
R0lGODlhEAAQAPUAMf////+Zmf+Sjv+HhP98ef9aV/9TUf9MSv8zM/8rKf8m
Jf8eHv8WFv8QD/8KCv8AAPYKCu4EA+MGBt4CAtgFBdMBAcQBAb4AAKoAAKcA
AJkAAJR8eY4AAIQAAICAgHwAAHIAAGYAAF4AACQeHhEODgAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAZx
QIBwCHgQj8iHBzk0NpdCJ9EoIBgSjCL0GAgMDIqGY0ssEAabw8gxuZCFCEPB
gGhMLJntY79IJBQMEhcYHHtSAA0NJBAUFyUcHx6SehETJZcdHyFvQhUXGRwd
ICCcWhqiISJaR0pPUK1DZLBMrKW0pUEAOw==
}]
 
set ::UI::icons(gadugadu_online) [image create photo -data {
R0lGODlhEAAQAPQAMf//////AP/3AP/vAP/nAP/eAP/WAP8AAPfWAO/OAOfG
AN7GAM61AM6tAL2lALWcAK2UAKWMAJyMAJyEAJSEAJR7AIx7AISEhAAAAAAA
AAAAAAAAAAAAAAAAAAAAAAAAACH5BAEAAAAALAAAAAAQABAAAAVbICCOwEGe
6HGho9muokuaxyHUJXweQSAUCRWLF8AQMAqGcOcbGBQNx7JUEwwIiEXjEZnU
ZKUCRvvAUFSXdCuxwLgpluksCqFU5K+aBIefomM6AIF4LICFLIEjIQA7
}]

# More

# General icon for minimal popup menu button.
set ::UI::icons(popupbt) [image create photo -data {
    R0lGODdhEAAOALMAAP///+/v797e3s7Ozr29va2trZycnJSUlIyMjHl5eXR0
    dHNzc2NjY1JSUkJCQgAAACwAAAAAEAAOAAAEYLDIIqoYg5A5iQhVthkGR4HY
    JhkI0gWgphUkspRn5ey8Yy+S0CDRcxRsjOAlUzwuGkERQcGjGZ6S1IxHWjCS
    hZkEcTC2vEAOieRDNlyrNevMaKQnrIXe+71zkF8MC3ASEQA7
}]
set ::UI::icons(popupbtpush) [image create photo -data {
    R0lGODdhEAAOAMQAAP///9DQ0M7Ozr29va2trZycnI2NjYyMjHZ2dnV1dXR0
    dHNzc29vb2NjY1JSUkJCQgAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAAAAAACwAAAAAEAAOAAAFbqAgCsTSOGhTjMLgmk0s
    O4s7EGW8HEVxHKYDjnCI8Vy4HnBoWhBsN+WigIsVBoGHdvtQCAm0qwDBfRAS
    TvDUVjYs0g6VjbEluL/WIWELnOKKPD0FdAQKbzc4b4EHBg9vUyRDP5N9kFGC
    UjtULT0hADs=
}]
    
# The mac look-alike triangles.
set ::UI::icons(mactriangleopen) [image create photo -data {
    R0lGODlhCwALAPMAAP///97e3s7O/729vZyc/4yMjGNjzgAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAEALAAAAAALAAsAAAQgMMhJq7316M1P
    OEIoEkchHURKGOUwoWubsYVryZiNVREAOw==
}]
set ::UI::icons(mactriangleclosed) [image create photo -data {
    R0lGODlhCwALAPMAAP///97e3s7O/729vZyc/4yMjGNjzgAAAAAAAAAAAAAA
    AAAAAAAAAAAAAAAAAAAAACH5BAEAAAEALAAAAAALAAsAAAQiMMgjqw2H3nqE
    3h3xWaEICgRhjBi6FgMpvDEpwuCBg3sVAQA7
}]

#-------------------------------------------------------------------------------
