# Tcl package index file, version 1.1
# This file is generated by the "pkg_mkIndex" command
# and sourced either when an application starts up or
# by a "package unknown" script.  It invokes the
# "package ifneeded" command to set up package-related
# information so that packages will be loaded automatically
# in response to "package require" commands.  When this
# script is sourced, the variable $dir must contain the
# full path name of this file's directory.

## The following function will load the tkdnd shared library only when
## a "package require tkdnd" command is issued, and not when tcl searches for
## packages. This bug was reported and fixed by 
## Colin McDonald <mcdonald@boscombe.demon.co.uk>
namespace eval ::dnd {
    proc tkdndLoad {dir} {
	load [file join $dir libtkdnd10[info sharedlibextension]]
	source [file join $dir tkDND_Utils.tcl]
	package provide tkdnd 1.0
    }
}

package ifneeded tkdnd 1.0 [list ::dnd::tkdndLoad $dir]
