# Tcl package index file, version 1.1
# This file is generated by the "pkg_mkIndex" command
# and sourced either when an application starts up or
# by a "package unknown" script.  It invokes the
# "package ifneeded" command to set up package-related
# information so that packages will be loaded automatically
# in response to "package require" commands.  When this
# script is sourced, the variable $dir must contain the
# full path name of this file's directory.

package ifneeded Agents 1.0 [list source [file join $dir Agents.tcl]]
package ifneeded Browse 1.0 [list source [file join $dir Browse.tcl]]
package ifneeded Chat 1.0 [list source [file join $dir Chat.tcl]]
package ifneeded Conference 1.0 [list source [file join $dir Conference.tcl]]
package ifneeded Disco 1.0 [list source [file join $dir Disco.tcl]]
package ifneeded Emoticons 1.0 [list source [file join $dir Emoticons.tcl]]
package ifneeded GotMsg 1.0 [list source [file join $dir GotMsg.tcl]]
package ifneeded GroupChat 1.0 [list source [file join $dir GroupChat.tcl]]
package ifneeded History 1.0 [list source [file join $dir History.tcl]]
package ifneeded Jabber 1.0 [list source [file join $dir Jabber.tcl]]
package ifneeded JForms 1.0 [list source [file join $dir JForms.tcl]]
package ifneeded JPrefs 1.0 [list source [file join $dir JPrefs.tcl]]
package ifneeded JUI 1.0 [list source [file join $dir JUI.tcl]]
package ifneeded JUser 1.0 [list source [file join $dir JUser.tcl]]
package ifneeded JWB 1.0 [list source [file join $dir JWB.tcl]]
package ifneeded Login 1.0 [list source [file join $dir Login.tcl]]
package ifneeded MailBox 1.0 [list source [file join $dir MailBox.tcl]]
package ifneeded MUC 1.0 [list source [file join $dir MUC.tcl]]
package ifneeded NewMsg 1.0 [list source [file join $dir NewMsg.tcl]]
package ifneeded OOB 1.0 [list source [file join $dir OOB.tcl]]
package ifneeded Privacy 1.0 [list source [file join $dir Privacy.tcl]]
package ifneeded Profiles 1.0 [list source [file join $dir Profiles.tcl]]
package ifneeded Register 1.0 [list source [file join $dir Register.tcl]]
package ifneeded Roster 1.0 [list source [file join $dir Roster.tcl]]
package ifneeded Search 1.0 [list source [file join $dir Search.tcl]]
package ifneeded SetupAss 1.0 [list source [file join $dir SetupAss.tcl]]
package ifneeded Subscribe 1.0 [list source [file join $dir Subscribe.tcl]]
package ifneeded VCard 1.0 [list source [file join $dir VCard.tcl]]

